--Brainfuck to PAP compiler
-- compiles Brainfuck programs down to PAP programs

program = {} --the reduced program
pap = {} --the pap program
heap_ptr_reg = 3 --the register which holds the heap pointer
temp_reg = 4 --the register which holds temporary values
loop_stack = {} --stack of locations to go back to (for looping)
loop_return_stack = {}

--function to process a file
function process_file(f)
  for line in io.lines(f) do
    for c = 1, #line, 1 do
      local char = string.sub(line, c, c)
      if string.match(char, "[%<%>%+%-%.%,%[%]]") then
        program[#program+1] = char
      end
    end
  end
end

--print out a table
function printTable(t)
  print ""
  for k,v in pairs(t) do
    print(k,v)
  end
end

--adds instruction to pap program
function add_instr(i)
  pap[#pap+1] = i
end

--compile function
function compile()
  for k, v in pairs(program) do
    if v == ">" then
      add_instr("2 1 2 "..heap_ptr_reg) --add the heap ptr and 1
      add_instr("1 3 "..heap_ptr_reg) --copy value back into the register
      add_instr("1 1") --pop
    elseif v == "<" then
      add_instr("2 2 "..heap_ptr_reg.." 2") --sub the heap ptr and 1
      add_instr("1 3 "..heap_ptr_reg) --copy value back into the register
      add_instr("1 1") --pop
    elseif v == "+" then
      add_instr("1 2 "..heap_ptr_reg) --push heap ptr onto stack
      add_instr("1 5") --push heap value
      add_instr("1 3 "..temp_reg) --copy value into temp reg
      add_instr("1 1") --pop val
      add_instr("2 1 2 "..temp_reg) --add one to value
      add_instr("1 2 "..heap_ptr_reg) --push heap ptr
      add_instr("1 6") --copy to heap
    elseif v == "-" then
      add_instr("1 2 "..heap_ptr_reg) --push heap ptr onto stack
      add_instr("1 5") --push heap value
      add_instr("1 3 "..temp_reg) --copy value into temp reg
      add_instr("1 1") --pop val
      add_instr("2 2 "..temp_reg.." 2") --sub one from value
      add_instr("1 2 "..heap_ptr_reg) --push heap ptr
      add_instr("1 6") --copy to heap
    elseif v == "." then
      add_instr("1 2 "..heap_ptr_reg) --push heap ptr onto stack
      add_instr("1 5") --push heap value
      add_instr("4 1") --print character
    elseif v == "," then
      add_instr("4 3") --read char
      add_instr("1 2 "..heap_ptr_reg) --push heap ptr onto stack
      add_instr("1 6") --copy value to heap
    elseif v == "[" then
      --prepare for comparisons
      loop_return_stack[#loop_return_stack+1] = #pap
      add_instr("1 2 "..heap_ptr_reg) --push heap ptr onto stack
      add_instr("1 5") --push heap value
      add_instr("1 3 "..temp_reg) --copy value into temp reg
      add_instr("1 1") --pop val
      
      add_instr("1 4 "..#pap+4) --push current loc + 4
      add_instr("3 1 1 "..temp_reg) --BNE on zero and temp_reg
      add_instr(" ") --push end loc + n (empty, calculate later)
      loop_stack[#loop_stack+1] = #pap
      add_instr("5")
    elseif v == "]" then
      add_instr("1 4 "..loop_return_stack[#loop_return_stack]) --push first line loc
      add_instr("5") --jump!
      
      --rewrite the appropriate line then get pop off stacks
      pap[loop_stack[#loop_stack]] = "1 4 "..#pap
      loop_stack[#loop_stack] = nil
      loop_return_stack[#loop_return_stack] = nil
    end
  end
end


--main function
function main()
  --process program into instruction store
  if arg[1] == nil then
    print "ERROR: No program specified"
    return
  else  
    process_file(arg[1])
  end

  --compile program
  compile()

  --write out program
  for k,v in pairs(pap) do
    print(v)
  end
  
  io.write("\n") --clean up output
end

--run
main()